import logging
from datetime import datetime
from typing import Iterable, Tuple

import pytz

from fetchfox.apis.algorand import (
    algonodecloud,
    nfdomains,
    randswapcom,
    algoxnftcom,
    nftexplorerapp,
)
from fetchfox.blockchains.base import Blockchain
from fetchfox.constants.blockchains import ALGORAND
from fetchfox.constants.currencies import ALGO
from fetchfox.constants.marketplaces import (
    ALGOXNFT_COM,
    RANDGALLERY_COM,
    SHUFL_APP,
    UNKNOWN,
)
from fetchfox.dtos import (
    AssetDTO,
    CampaignDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    SaleDTO,
)
from . import utils
from .exceptions import (
    InvalidAlgorandAssetIdException,
    InvalidAlgorandCollectionIdException,
    InvalidAlgorandWalletException,
)

logger = logging.getLogger(__name__)


class Algorand(Blockchain):
    def __init__(self, nftexplorerapp_api_key: str = None):
        super().__init__(
            name=ALGORAND,
            currency=ALGO,
            logo="https://s2.coinmarketcap.com/static/img/coins/64x64/4030.png",
        )

        self.nftexplorerapp_api_key: str = nftexplorerapp_api_key

    def check_asset_id(self, asset_id: str):
        if not utils.is_asset_id(asset_id):
            raise InvalidAlgorandAssetIdException(asset_id)

    def check_collection_id(self, collection_id: str):
        if not utils.is_address(collection_id):
            raise InvalidAlgorandCollectionIdException(collection_id)

    def check_wallet(self, wallet: str):
        if not utils.is_wallet(wallet):
            raise InvalidAlgorandWalletException(wallet)

    def explorer_url(self, *, address: str = None, collection_id: str = None, asset_id: str = None, tx_hash: str = None) -> str:
        if address:
            return f"https://algoexplorer.io/address/{address.upper()}"

        if asset_id:
            return f"https://algoexplorer.io/asset/{asset_id}"

        if collection_id:
            return f"https://algoexplorer.io/address/{collection_id.upper()}"

        if tx_hash:
            return f"https://algoexplorer.io/tx/{tx_hash.upper()}"

        return None

    def marketplace_url(self, *, collection_id: str = None, asset_id: str = None) -> str:
        if asset_id:
            return f"https://www.randgallery.com/algo-collection/?address={asset_id}"

        if collection_id:
            return f"https://randgallery.com/algo-collection/?address={collection_id}"

        return None

    def get_wallet_name(self, wallet: str) -> str:
        if utils.is_nf_domain(wallet):
            return wallet

        if utils.is_address(wallet):
            return nfdomains.get_nf_domain(wallet)

        return None

    def resolve_wallet_name(self, wallet_name: str) -> str:
        if utils.is_nf_domain(wallet_name):
            return nfdomains.resolve_nf_domain(wallet_name)

        return None

    def get_asset(self, collection_id: str, asset_id: str, fetch_metadata: bool = True, *args, **kwargs) -> AssetDTO:
        if collection_id:
            self.check_collection_id(collection_id)

        self.check_asset_id(asset_id)

        data = algonodecloud.get_asset_data(asset_id)

        if fetch_metadata:
            metadata = algonodecloud.get_asset_metadata(asset_id)
            metadata["name"] = data["name"]
        else:
            metadata = {}

        return AssetDTO(
            collection_id=data["creator"],
            asset_id=asset_id,
            metadata=metadata,
        )

    def get_assets(self, collection_id: str, fetch_metadata: bool = True, *args, **kwargs) -> Iterable[AssetDTO]:
        self.check_collection_id(collection_id)

        for asset_id in algonodecloud.get_assets(collection_id):
            if fetch_metadata:
                yield self.get_asset(
                    collection_id=collection_id,
                    asset_id=str(asset_id),
                )
            else:
                yield AssetDTO(
                    collection_id=collection_id,
                    asset_id=str(asset_id),
                    metadata={},
                )

    def get_balance(self, wallet: str) -> Tuple[float, str]:
        self.check_wallet(wallet)

        if utils.is_nf_domain(wallet):
            wallet = nfdomains.resolve_nf_domain(wallet)

        balance = algonodecloud.get_balance(wallet)

        return balance, self.currency

    def get_campaigns(self) -> Iterable[CampaignDTO]:
        return

    def get_floor(self, collection_id: str, *args, **kwargs) -> FloorDTO:
        self.check_collection_id(collection_id)

        floor = None
        count = 0

        for listing in self.get_listings(collection_id):
            count += 1

            if floor is None:
                floor = listing
            elif listing.usd < floor.usd:
                floor = listing

        return FloorDTO(
            listing=floor,
            listing_count=count,
        )

    def get_holdings(self, wallet: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_wallet(wallet)

        if utils.is_nf_domain(wallet):
            wallet = nfdomains.resolve_nf_domain(wallet)

        holdings = algonodecloud.get_holdings(wallet)

        for holding in holdings:
            asset_id = holding["asset-id"]
            quantity = holding["amount"]

            if quantity < 1:
                continue

            yield HoldingDTO(
                collection_id=None,
                asset_id=asset_id,
                address=wallet,
                quantity=quantity,
            )

    def get_listings(self, collection_id: str, *args, **kwargs) -> Iterable[ListingDTO]:
        self.check_collection_id(collection_id)

        listings = randswapcom.get_listings(collection_id)

        for listing in listings:
            asset_id = str(listing["assetId"])
            asset_ids = [asset_id]
            asset_names = [""]

            listed_at = datetime.fromtimestamp(
                listing["timestamp"] // 1000,
            ).replace(
                tzinfo=pytz.UTC,
            )

            yield ListingDTO(
                identifier=listing["timestamp"],
                collection_id=collection_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                listing_id=listing["timestamp"],
                marketplace=RANDGALLERY_COM,
                price=listing["price"],
                currency=ALGO,
                listed_at=listed_at,
                listed_by=listing["sellerAddress"],
                marketplace_url=f"https://randgallery.com/algo-collection/?address={asset_id}",
            )

        listings = algoxnftcom.get_listings(collection_id)

        for listing in listings:
            asset_id = str(listing["asset_id"])
            asset_ids = [asset_id]
            asset_names = [""]

            listed_at = datetime.now(tz=pytz.utc)

            yield ListingDTO(
                identifier=listing["buy_it_now_listing_id"],
                collection_id=collection_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                listing_id=listing["buy_it_now_listing_id"],
                marketplace=ALGOXNFT_COM,
                price=listing["price"] // 10**6,
                currency=self.currency,
                listed_at=listed_at,
                listed_by=listing["seller"],
                marketplace_url=f"https://algoxnft.com/asset/{asset_id}",
            )

    def get_owners(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        holding = algonodecloud.get_owner(str(asset_id))

        yield HoldingDTO(
            collection_id=collection_id,
            asset_id=holding["asset_id"],
            address=holding["address"],
            quantity=holding["amount"],
        )

    def get_sales(self, collection_id: str, *args, **kwargs) -> Iterable[SaleDTO]:
        self.check_collection_id(collection_id)

        venues = {
            "algoxnft": ALGOXNFT_COM,
            "randgallery": RANDGALLERY_COM,
            "shufl": SHUFL_APP,
            None: UNKNOWN,
        }

        sales = nftexplorerapp.get_sales(
            collection_id,
            api_key=self.nftexplorerapp_api_key,
        )

        for sale in sales:
            tx_hash = sale["txnId"]

            asset_id = str(sale["asset"])
            asset_name = sale["assetName"]

            confirmed_at = datetime.fromtimestamp(
                sale["epochMs"] // 1000,
            ).replace(
                tzinfo=pytz.UTC,
            )

            marketplace = venues[sale.get("venue")]

            if marketplace == RANDGALLERY_COM:
                marketplace_url = f"https://randgallery.com/algo-collection/?address={asset_id}"
            elif marketplace == ALGOXNFT_COM:
                marketplace_url = f"https://algoxnft.com/asset/{asset_id}"
            elif marketplace == SHUFL_APP:
                marketplace_url = f"https://shufl.app/detail/{asset_id}"
            else:
                marketplace_url = None

            yield SaleDTO(
                identifier=tx_hash,
                collection_id=collection_id,
                asset_ids=[asset_id],
                asset_names=[asset_name],
                tx_hash=tx_hash,
                marketplace=marketplace,
                price=sale["ualgos"] // 10**6,
                currency=self.currency,
                confirmed_at=confirmed_at,
                sold_by=sale["sender"],
                bought_by=sale["receiver"],
                marketplace_url=marketplace_url,
                explorer_url=f"https://algoexplorer.io/tx/{tx_hash}",
            )

    def get_snapshot(self, collection_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)

        for asset in self.get_assets(collection_id, fetch_metadata=False):
            yield from self.get_owners(collection_id, asset.asset_id)



from abc import abstractmethod
from typing import Iterable, Tuple

from fetchfox.apis import coingeckocom
from fetchfox.dtos import (
    AssetDTO,
    CampaignDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    RankDTO,
    SaleDTO,
)


class Blockchain:
    def __init__(self, name: str, currency: str, logo: str):
        self.name: str = name
        self.currency: str = currency
        self.logo: str = logo

    @property
    def usd(self) -> float:
        return coingeckocom.usd(self.currency)

    @abstractmethod
    def check_collection_id(self, collection_id: str):
        raise NotImplementedError()

    @abstractmethod
    def check_asset_id(self, asset_id: str):
        raise NotImplementedError()

    @abstractmethod
    def check_wallet(self, wallet: str):
        raise NotImplementedError()

    @abstractmethod
    def explorer_url(self, *, address: str = None, collection_id: str = None, asset_id: str = None, tx_hash: str = None) -> str:
        raise NotImplementedError()

    @abstractmethod
    def marketplace_url(self, *, collection_id: str = None, asset_id: str = None) -> str:
        raise NotImplementedError()

    @abstractmethod
    def get_wallet_name(self, wallet: str) -> str:
        raise NotImplementedError()

    @abstractmethod
    def resolve_wallet_name(self, wallet: str) -> str:
        raise NotImplementedError()

    def format_wallet(self, wallet: str) -> str:
        self.check_wallet(wallet)

        name = self.get_wallet_name(wallet)

        if name:
            return name

        return f"{wallet[:5]}..{wallet[-5:]}"

    @abstractmethod
    def get_asset(self, collection_id: str, asset_id: str, *args, **kwargs) -> AssetDTO:
        raise NotImplementedError()

    @abstractmethod
    def get_assets(self, collection_id: str, fetch_metadata: bool = True, *args, **kwargs) -> Iterable[AssetDTO]:
        raise NotImplementedError()

    @abstractmethod
    def get_balance(self, wallet: str) -> Tuple[float, str]:
        raise NotImplementedError()

    def get_campaigns(self) -> Iterable[CampaignDTO]:
        raise NotImplementedError()

    @abstractmethod
    def get_floor(self, collection_id: str, *args, **kwargs) -> FloorDTO:
        raise NotImplementedError()

    @abstractmethod
    def get_holdings(self, wallet: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        raise NotImplementedError()

    @abstractmethod
    def get_listings(self, collection_id: str, *args, **kwargs) -> Iterable[ListingDTO]:
        raise NotImplementedError()

    @abstractmethod
    def get_owners(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        raise NotImplementedError()

    def get_rank(self, collection_id: str, asset_id: str, *args, **kwargs) -> RankDTO:
        return None

    def get_ranks(self, collection_id: str, *args, **kwargs) -> Iterable[RankDTO]:
        return []

    @abstractmethod
    def get_snapshot(self, collection_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        raise NotImplementedError()

    @abstractmethod
    def get_sales(self, collection_id: str, *args, **kwargs) -> Iterable[SaleDTO]:
        raise NotImplementedError()



from fetchfox.apis import coingeckocom
from fetchfox.apis.cardano import muesliswap
from fetchfox.constants.book import (
    BOOK_TOKEN_ASSET_ID,
    BOOK_TOKEN_ASSET_NAME,
    BOOK_TOKEN_POLICY_ID,
    BOOK_TOKEN_FINGERPRINT,
    BOOK_TOKEN_SYMBOL,
)
symbol = BOOK_TOKEN_SYMBOL
policy_id = BOOK_TOKEN_POLICY_ID
asset_name = BOOK_TOKEN_ASSET_NAME
asset_id = BOOK_TOKEN_ASSET_ID
fingetprint = BOOK_TOKEN_FINGERPRINT

cardanoscan_url = f"https://cardanoscan.io/token/{asset_id}"
cexplorer_url = f"https://cexplorer.io/asset/{fingetprint}"
minswap_url = f"https://app.minswap.org/swap?currencySymbolA={policy_id}&tokenNameA={asset_name}&currencySymbolB=&tokenNameB="


def ada() -> float:
    return muesliswap.price(asset_id)


def usd() -> float:
    from fetchfox.constants.currencies import ADA

    return ada() * coingeckocom.usd(ADA)



import json
import logging
from datetime import datetime
from typing import Iterable, Tuple

import pytz

from fetchfox.apis import bookio
from fetchfox.apis.evm import moralisio, ensideascom, openseaio
from fetchfox.blockchains.base import Blockchain
from fetchfox.constants.blockchains import ETHEREUM, POLYGON
from fetchfox.constants.marketplaces import OPENSEA_IO
from fetchfox.dtos import AssetDTO, CampaignDTO, FloorDTO, HoldingDTO, ListingDTO, SaleDTO
from fetchfox.helpers import formatters
from . import utils
from .exceptions import (
    InvalidEvmAssetIdException,
    InvalidEvmCollectionIdException,
    InvalidEvmWalletException,
)

logger = logging.getLogger(__name__)


class Evm(Blockchain):
    def __init__(self, name: str, currency: str, logo: str, moralisio_api_key: str = None, openseaio_api_key: str = None):
        super().__init__(name, currency, logo)

        self.moralisio_api_key: str = moralisio_api_key
        self.openseaio_api_key: str = openseaio_api_key

    def check_asset_id(self, asset_id: str):
        if not utils.is_asset_id(asset_id):
            raise InvalidEvmAssetIdException(asset_id, self.name)

    def check_collection_id(self, collection_id: str):
        if not utils.is_address(collection_id):
            raise InvalidEvmCollectionIdException(collection_id, self.name)

    def check_wallet(self, wallet: str):
        if not utils.is_wallet(wallet):
            raise InvalidEvmWalletException(wallet, self.name)

    def get_wallet_name(self, wallet: str) -> str:
        if utils.is_ens_domain(wallet):
            return wallet

        if utils.is_address(wallet):
            return ensideascom.get_ens_domain(wallet)

        return None

    def resolve_wallet_name(self, wallet_name: str) -> str:
        if utils.is_ens_domain(wallet_name):
            return ensideascom.resolve_ens_domain(wallet_name)

        return None

    def get_asset(self, collection_id: str, asset_id: str, *args, **kwargs) -> AssetDTO:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        data = moralisio.get_asset_data(
            contract_address=collection_id,
            asset_id=asset_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
        )

        metadata = json.loads(data["metadata"])
        metadata["attributes"].update(metadata["extraAttributes"])

        return AssetDTO(
            collection_id=collection_id,
            asset_id=asset_id,
            metadata=metadata,
        )

    def get_assets(self, collection_id: str, fetch_metadata: bool = True, *args, **kwargs) -> Iterable[AssetDTO]:
        self.check_collection_id(collection_id)

        if fetch_metadata:
            asset_id = -1

            while True:
                try:
                    asset_id += 1

                    yield self.get_asset(
                        collection_id=collection_id,
                        asset_id=str(asset_id),
                    )
                except ValueError:
                    raise
                except:
                    break
        else:
            assets = moralisio.get_assets(
                contract_address=collection_id,
                blockchain=self.name,
                api_key=self.moralisio_api_key,
            )

            for asset_id in assets:
                yield AssetDTO(
                    collection_id=collection_id,
                    asset_id=asset_id,
                    metadata={},
                )

    def get_balance(self, wallet: str) -> Tuple[float, str]:
        self.check_wallet(wallet)

        if utils.is_ens_domain(wallet):
            wallet = ensideascom.resolve_ens_domain(wallet)

        balance = moralisio.get_balance(
            wallet,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
        )

        return balance, self.currency

    def get_campaigns(self) -> Iterable[CampaignDTO]:
        for campaign in bookio.campaigns():
            if campaign["blockchain"] != "evm":
                continue

            if self.name == ETHEREUM:
                network = "mainnet"
            else:
                network = self.name

            if campaign["network"] != network:
                continue

            landing = campaign["landing"]
            platform = campaign["platform"]
            start_at = campaign["start_at"]

            collection_id = platform["ThirdWeb"]["contract_address"]

            yield CampaignDTO(
                blockchain=self.name,
                parlamint_id=campaign["id"],
                collection_id=collection_id,
                name=campaign["display_name"],
                description=campaign["landing"]["opener"],
                price=landing["price"],
                supply=campaign["total_deas"],
                pricing=landing["price_description"],
                limit=campaign["max_quantity"],
                start_at=formatters.timestamp(start_at["start_at"]) if start_at["type"] == "at" else None,
                explorer_url=self.explorer_url(collection_id=collection_id),
                marketplace_url=self.marketplace_url(collection_id=collection_id),
                rarity_chart_url=landing["rarity_chart_url"],
                animated_cover_url=landing["animated_cover_url"],
            )

    def get_floor(self, collection_id: str, *args, **kwargs) -> FloorDTO:
        self.check_collection_id(collection_id)

        floor = None
        count = 0

        for listing in self.get_listings(collection_id):
            count += 1

            if floor is None:
                floor = listing
            elif listing.usd < floor.usd:
                floor = listing

        return FloorDTO(
            listing=floor,
            listing_count=count,
        )

    def get_holdings(self, wallet: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_wallet(wallet)

        if utils.is_ens_domain(wallet):
            wallet = ensideascom.resolve_ens_domain(wallet)

        holdings = moralisio.get_holdings(
            wallet,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
        )

        for token in holdings:
            contract_address = token["token_address"]
            asset_id = token["token_id"]
            quantity = int(token["amount"])

            yield HoldingDTO(
                collection_id=contract_address,
                asset_id=asset_id,
                address=wallet,
                quantity=quantity,
            )

    def get_listings(self, collection_id: str, slug: str = None, *args, **kwargs) -> Iterable[ListingDTO]:
        self.check_collection_id(collection_id)

        listings = openseaio.get_listings(
            collection_id,
            api_key=self.openseaio_api_key,
            slug=slug,
        )

        for listing in listings:
            protocol_data = listing["protocol_data"]
            parameters = protocol_data["parameters"]

            price = listing["price"]["current"]

            asset_ids = []
            asset_names = []

            for offer in parameters["offer"]:
                if offer["token"].lower() != collection_id.lower():
                    continue

                asset_ids.append(offer["identifierOrCriteria"])
                asset_names.append("")

            if listing.get("eventTimestamp"):
                listed_at = datetime.utcfromtimestamp(parameters["startTime"])
            else:
                listed_at = datetime.now(tz=pytz.utc)

            if self.name == POLYGON:
                marketplace_url = f"https://opensea.io/assets/matic/{collection_id}/{asset_ids[0]}"
            else:
                marketplace_url = f"https://opensea.io/assets/ethereum/{collection_id}/{asset_ids[0]}"

            yield ListingDTO(
                identifier=listing["order_hash"],
                collection_id=collection_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                listing_id=listing["order_hash"],
                marketplace=OPENSEA_IO,
                price=float(int(price["value"]) / 10 ** price["decimals"]),
                currency=price["currency"].replace("WETH", "ETH"),
                listed_at=listed_at,
                listed_by=parameters["offerer"],
                tx_hash=listing["order_hash"],
                marketplace_url=marketplace_url,
            )

    def get_owners(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset = moralisio.get_owner(
            collection_id,
            asset_id=asset_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
        )

        yield HoldingDTO(
            collection_id=asset["contract_address"],
            asset_id=asset["asset_id"],
            address=asset["address"],
            quantity=asset["amount"],
        )

    def get_sales(self, collection_id: str, slug: str = None, *args, **kwargs) -> Iterable[SaleDTO]:
        self.check_collection_id(collection_id)

        sales = openseaio.get_sales(
            collection_id,
            api_key=self.openseaio_api_key,
            slug=slug,
        )

        for sale in sales:
            tx_hash = sale["transaction"]["transaction_hash"]

            asset = sale["asset"]
            asset_id = asset["token_id"]
            asset_name = asset["name"]

            if sale.get("event_timestamp"):
                confirmed_at = datetime.fromisoformat(sale["event_timestamp"] + "+00:00")
            else:
                confirmed_at = datetime.now(tz=pytz.utc)

            if self.name == POLYGON:
                marketplace_url = f"https://opensea.io/assets/matic/{collection_id}/{asset_id}"
                explorer_url = f"https://polygonscan.com/tx/{tx_hash}"
            else:
                marketplace_url = f"https://opensea.io/assets/ethereum/{collection_id}/{asset_id}"
                explorer_url = f"https://etherscan.io/tx/{tx_hash}"

            yield SaleDTO(
                identifier=sale["id"],
                collection_id=collection_id,
                asset_ids=[asset_id],
                asset_names=[asset_name],
                tx_hash=tx_hash,
                marketplace=OPENSEA_IO,
                price=float(int(sale["total_price"]) / 10 ** sale["payment_token"]["decimals"]),
                currency=sale["payment_token"]["symbol"].replace("WETH", "ETH"),
                confirmed_at=confirmed_at,
                sold_by=sale["seller"]["address"],
                bought_by=sale["winner_account"]["address"],
                sale_id=sale["id"],
                marketplace_url=marketplace_url,
                explorer_url=explorer_url,
            )

    def get_snapshot(self, collection_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)

        assets = moralisio.get_owners(
            collection_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
        )

        for asset in assets:
            yield HoldingDTO(
                collection_id=asset["contract_address"],
                asset_id=asset["asset_id"],
                address=asset["address"],
                quantity=asset["amount"],
            )



import base64
from functools import lru_cache
from typing import Iterable, Tuple

from fetchfox import rest
from fetchfox.checks import check_str
from fetchfox.constants.cardano import ADA_HANDLE_POLICY_ID

BASE_URL = "https://cardano-mainnet.blockfrost.io/api"


def get(service: str, project_id: str, params: dict = None, version: int = 0) -> Tuple[dict, int]:
    check_str(project_id, "blockfrost.project_id")

    return rest.get(
        url=f"{BASE_URL}/v{version}/{service}",
        params=params or {},
        headers={
            "project_id": project_id,
        },
    )


def get_assets(policy_id: str, project_id: str) -> Iterable[str]:
    check_str(policy_id, "blockfrost.policy_id")
    policy_id = policy_id.strip().lower()

    page = 0

    while True:
        page += 1

        response, status_code = get(
            f"assets/policy/{policy_id}",
            params={
                "page": page,
                "count": 100,
                "order": "desc",
            },
            project_id=project_id,
        )

        if not response:
            break

        for item in response:
            if int(item["quantity"]) == 0:
                continue

            if item["asset"] == policy_id:
                continue

            yield item["asset"]


@lru_cache(maxsize=None)
def get_asset_data(asset_id: str, project_id: str) -> dict:
    check_str(asset_id, "blockfrost.asset_id")

    asset_id = asset_id.strip().lower()

    response, status_code = get(
        f"assets/{asset_id}",
        project_id=project_id,
    )

    return response


@lru_cache(maxsize=None)
def get_stake_address(address: str, project_id: str) -> str:
    check_str(address, "blockfrost.address")

    response, status_code = get(
        f"addresses/{address}",
        project_id=project_id,
    )

    return response.get("stake_address")


def get_balance(stake_address: str, project_id: str) -> float:
    check_str(stake_address, "blockfrost.stake_address")

    response, status_code = get(
        f"accounts/{stake_address}",
        project_id=project_id,
    )

    return int(response["controlled_amount"]) / 10**6


def get_holdings(stake_address: str, project_id: str) -> Iterable[dict]:
    check_str(stake_address, "blockfrost.stake_address")

    page = 0

    while True:
        page += 1

        response, status_code = get(
            f"accounts/{stake_address}/addresses/assets",
            params={
                "count": 100,
                "page": page,
            },
            project_id=project_id,
        )

        if not response:
            break

        yield from response


def get_owners(asset_id: str, project_id: str) -> dict:
    check_str(asset_id, "blockfrost.asset_id")

    page = 0

    while True:
        page += 1
        response, status_code = get(
            f"assets/{asset_id}/addresses",
            params={
                "page": page,
            },
            project_id=project_id,
        )

        if isinstance(response, dict):
            if response.get("error"):
                return None

        if not response:
            break

        for item in response:
            if item.get("quantity") == "0":
                continue

            address = item["address"]
            stake_address = get_stake_address(address, project_id)

            yield {
                "asset_id": asset_id,
                "address": address,
                "stake_address": stake_address,
                "amount": int(item["quantity"]),
            }


@lru_cache(maxsize=None)
def get_handle(stake_address: str, project_id: str) -> str:
    handles = []

    for holding in get_holdings(stake_address, project_id=project_id):
        if holding["quantity"] != "1":
            continue

        asset_id = holding["unit"]

        if not asset_id.startswith(ADA_HANDLE_POLICY_ID):
            continue

        asset_id = asset_id.replace(f"{ADA_HANDLE_POLICY_ID}000de140", "")  # CIP-68
        asset_id = asset_id.replace(ADA_HANDLE_POLICY_ID, "")  # CIP-25

        asset_name = bytes.fromhex(asset_id).decode()
        handles.append(asset_name)

    if not handles:
        return None

    return sorted(handles, key=len)[0]


@lru_cache(maxsize=None)
def resolve_handle(handle: str, project_id: str) -> str:
    check_str(handle, "blockfrost.handle")

    if handle.startswith("$"):
        if handle.startswith("$"):
            handle = handle[1:]

    handle = handle.lower()

    wallet = resolve_cip25_handle(handle, project_id)

    if wallet:
        return wallet

    return resolve_cip68_handle(handle, project_id)


def resolve_cip25_handle(handle: str, project_id: str) -> str:
    check_str(handle, "blockfrost.handle")

    encoded_name = base64.b16encode(handle.encode()).decode("utf-8")

    asset_id = f"{ADA_HANDLE_POLICY_ID}{encoded_name}".lower()
    owners = list(get_owners(asset_id, project_id))

    if not owners:
        return None

    owner = owners[0]
    owner["cip"] = 25

    return owner


def resolve_cip68_handle(handle: str, project_id: str) -> str:
    check_str(handle, "blockfrost.handle")

    encoded_name = base64.b16encode(handle.encode()).decode("utf-8")

    asset_id = f"{ADA_HANDLE_POLICY_ID}000de140{encoded_name}".lower()
    owners = list(get_owners(asset_id, project_id))

    if not owners:
        return None

    owner = owners[0]
    owner["cip"] = 68

    return owner



from functools import lru_cache
from typing import Tuple

from fetchfox import rest
from fetchfox.blockchains.cardano import utils

BASE_URL = "https://api.cnft.tools/api"


def get(service: str, params: dict = None) -> Tuple[dict, int]:
    return rest.get(
        url=f"{BASE_URL}/{service}",
        params=params or {},
    )


@lru_cache(maxsize=None)
def get_ranks(policy_id: str) -> dict:
    response, status_code = get(f"rankings/{policy_id}")

    if response.get("error"):
        return None

    return response


@lru_cache(maxsize=None)
def get_rank(asset_id: str) -> int:
    policy_id, asset_name = utils.split_asset_id(asset_id)
    asset_name = asset_name.replace(" ", "").replace("#", "")

    ranks = get_ranks(policy_id)

    if ranks is None:
        return None

    return ranks.get(asset_name)



BOOK_TOKEN_POLICY_ID = "51a5e236c4de3af2b8020442e2a26f454fda3b04cb621c1294a0ef34"
BOOK_TOKEN_ASSET_NAME = "424f4f4b"
BOOK_TOKEN_ASSET_ID = f"{BOOK_TOKEN_POLICY_ID}{BOOK_TOKEN_ASSET_NAME}"
BOOK_TOKEN_FINGERPRINT = "asset1lpmhzq33vsj9jpj2wsluu0usqfaf6kk96zdw5c"
BOOK_TOKEN_SYMBOL = "BOOK"
